@@ -12,12 +12,11 @@ require Exporter;
 @EXPORT_OK = qw(pd_raw %mtable %umult %wdays);
 
 use strict;
-use integer;
 
 # constants
 use vars qw(%mtable %umult %wdays $VERSION);
 
-$VERSION = 98.11_29_01;
+$VERSION = 99.06_23_01;
 
 # globals
 use vars qw($debug); 
@@ -25,6 +24,11 @@ use vars qw($debug);
 # dynamically-scoped
 use vars qw($parse);
 
+my %mtable;
+my %umult;
+my %wdays;
+my $y2k;
+
 CONFIG:	{
 
 	%mtable = qw(
@@ -55,6 +59,8 @@ CONFIG:	{
 		fri 5 friday 5
 		sat 6 saturday 6
 		);
+
+	$y2k = 946684800; # turn of the century
 }
 
 sub parsedate
@@ -72,7 +78,7 @@ sub parsedate
 	my $isspec;
 	my $now = $options{NOW} || time;
 	my $passes = 0;
-	my $uk = defined($options{UK})?$options{UK}:0;
+	my $uk = defined($options{UK}) ? $options{UK} : 0;
 
 	local $parse = '';  # will be dynamically scoped.
 
@@ -121,7 +127,7 @@ sub parsedate
 			}
 			next if $passes == 0 and $options{'TIMEFIRST'};
 			if (! defined $y) {
-				if (&parse_year_only(\$t, \$y)) {
+				if (&parse_year_only(\$t, \$y, $now, %options)) {
 					$parse .= " ".__LINE__ if $debug;
 					next;
 				}
@@ -165,6 +171,8 @@ sub parsedate
 
 		if ($passes == 0) {
 			print "nothing matched\n" if $debug;
+			return (undef, "no match on time/date") 
+				if wantarray();
 			return undef;
 		}
 	}
@@ -193,6 +201,8 @@ sub parsedate
 		# we didn't manage to eat the string
 		print "NOT WHOLE\n" if $debug;
 		return undef if $options{WHOLE};
+		return (undef, "characters left over after parse")
+			if wantarray();
 	}
 
 	# define a date if there isn't one already
@@ -201,20 +211,31 @@ sub parsedate
 		print "no date defined, trying to find one." if $debug;
 		if (defined $rs or defined $H) {
 			# we do have a time.
+			return (undef, "no date specified")
+				if wantarray();
 			return undef if $options{DATE_REQUIRED};
 			if (defined $rs) {
 				print "simple offset: $rs\n" if $debug;
-				return $now + $rs 
+				my $rv = $now + $rs;
+				return ($rv, $t) if wantarray();
+				return $rv;
 			}
 			$rd = 0;
 		} else {
 			print "no time either!\n" if $debug;
+			return (undef, "no time specified")
+				if wantarray();
 			return undef;
 		}
 	}
 
-	return undef if $options{TIME_REQUIRED} && ! defined($rs) 
-		&& ! defined($H) && ! defined($rd);
+	if ($options{TIME_REQUIRED} && ! defined($rs) 
+		&& ! defined($H) && ! defined($rd))
+	{
+		return (undef, "no time found")
+			if wantarray();
+		return undef;
+	}
 
 	my $secs;
 	my $jd;
@@ -233,7 +254,12 @@ sub parsedate
 			#
 			$isdst_now = (localtime($r))[8];
 			$isdst_then = (localtime($now))[8];
-			return $r if ($isdst_now == $isdst_then) || $options{GMT};
+			if (($isdst_now == $isdst_then) || $options{GMT})
+			{
+				return ($r, $t) if wantarray();
+				return $r 
+			}
+				
 			print "localtime changed DST during time period!\n" if $debug;
 		}
 
@@ -263,12 +289,20 @@ sub parsedate
 			$y += 1900;
 		}
 
-		if ($y < 100) {
-			# this will stop working at year 2100
-			$y += 100 if $y < 70;
-			$y += 1900;
-		}
+		$y = expand_two_digit_year($y, $now, %options)
+			if $y < 100;
 
+		if ($options{VALIDATE}) {
+			require Time::DaysInMonth;
+			my $dim = Time::DaysInMonth::days_in($y, $m);
+			if ($y < 1000 or $m < 1 or $d < 1 
+				or $y > 9999 or $m > 12 or $d > $dim)
+			{
+				return (undef, "illegal YMD: $y, $m, $d")
+					if wantarray();
+				return undef;
+			}
+		}
 		$jd = julian_day($y, $m, $d);
 		print "jd($y, $m, $d) = $jd\n" if $debug;
 	}
@@ -284,7 +318,9 @@ sub parsedate
 
 	my $carry;
 
-	print "before $rs $jd $H $M $S\n" if $debug;
+	print "before ", (defined($rs) ? "$rs" : ""),
+		    " $jd $H $M $S\n" 
+		if $debug;
 	#
 	# add in relative seconds.  Do it this way because we want to
 	# preserve the localtime across DST changes.
@@ -294,9 +330,19 @@ sub parsedate
 	$M = 0 unless $M; # -w
 	$H = 0 unless $H; # -w
 
+	if ($options{VALIDATE} and
+		($S < 0 or $M < 0 or $H > 0 or $S > 59 or $M > 59 or $H > 23)) 
+	{
+		return (undef, "illegal HMS: $H, $M, $S") if wantarray();
+		return undef;
+	}
+
 	$S += $rs if defined $rs;
 	$carry = int($S / 60);
+	my($frac) = $S - int($S);
+	$S = int($S);
 	$S %= 60;
+	$S += $frac;
 	$M += $carry;
 	$carry = int($M / 60);
 	$M %= 60;
@@ -347,6 +393,7 @@ sub parsedate
 
 	print "returning $secs.\n" if $debug;
 
+	return ($secs, $t) if wantarray();
 	return $secs;
 }
 
@@ -640,23 +687,24 @@ sub parse_time_only
 						(?:
 							\:
 							(\d\d)	(?# $7)
-								(?:
-									(?# don't barf on sybase millisecond timings)
-									\:\d\d\d
-								)?
+								(
+									(?# don't barf on database sub-second timings)
+									(?:\:|\.)
+									\d{1,6}
+								)?	(?# $8)
 						)?
 					)
 					\s*
-					([apAP][mM])?		(?# $8)
+					([apAP][mM])?		(?# $9)
 				) | (?:
-					(\d{1,2})		(?# $9)
-					([apAP][mM])		(?# ${10})
+					(\d{1,2})		(?# $10)
+					([apAP][mM])		(?# ${11})
 				)
 			)
 			(?:
 				\s+
 				"?
-				(				(?# ${11})
+				(				(?# ${12})
 					(?: [A-Z]{1,4}[TCW56] )
 					|
 					IDLE
@@ -670,11 +718,17 @@ sub parse_time_only
 			!!) { #"emacs
 		# HH[[:]MM[:SS]]meridan [zone] 
 		my $ampm;
-		$$hr = $1 || $5 || $9 || 0; # 9 is undef, but 5 is defined..
+		$$hr = $1 || $5 || $10 || 0; # 10 is undef, but 5 is defined..
 		$$mr = $2 || $6 || 0;
 		$$sr = $3 || $7 || 0;
-		$ampm = $4 || $8 || $10;
-		$$tzr = $11;
+		if (defined($8) && exists($options{SUBSECOND}) && $options{SUBSECOND}) {
+			my($frac) = $8;
+			substr($frac,0,1) = '.';
+			$$sr += $frac;
+		}
+		print "S = $$sr\n" if $debug;
+		$ampm = $4 || $9 || $11;
+		$$tzr = $12;
 		$$hr += 12 if $ampm and "\U$ampm" eq "PM" && $$hr != 12;
 		$$hr = 0 if $$hr == 12 && "\U$ampm" eq "AM";
 		$$hr = 0 if $$hr == 24;
@@ -723,6 +777,59 @@ sub parse_time_offset
 	return 0;
 }
 
+#
+# What to you do with a date that has a two-digit year?
+# There's not much that can be done except make a guess.
+#
+# Some example situations to handle:
+#
+#	now		year 
+#
+#	1999		01
+#	1999		71
+#	2010		71
+#	2110		09
+#
+
+sub expand_two_digit_year
+{
+	my ($yr, $now, %options) = @_;
+
+	return $yr if $yr > 100;
+
+	my ($y) = (&righttime($now, %options))[5];
+	$y += 1900;
+	my $century = int($y / 100) * 100;
+	my $within = $y % 100;
+
+	my $r = $yr + $century;
+
+	if ($options{PREFER_PAST}) {
+		if ($yr > $within) {
+			$r = $yr + $century - 100;
+		}
+	} elsif ($options{PREFER_FUTURE}) {
+		# being strict here would be silly
+		if ($yr < $within+10) {
+			# it's 2019 and the date is '08'
+			$r = $yr + $century + 100;
+		}
+	} else {
+		# prefer the current century in most cases
+
+		if ($within > 80 && $within - $yr > 60) {
+			$r = $yr + $century + 100;
+		}
+
+		if ($within < 30 && $yr - $within > 59) {
+			$r = $yr + $century - 100;
+		}
+	}
+	print "two digit year '$yr' expanded into $r\n" if $debug;
+	return $r;
+}
+
+
 sub calc 
 {
 	my ($rsr, $yr, $mr, $dr, $rdr, $now, $units, $count, %options) = @_;
@@ -752,7 +859,7 @@ sub monthoff
 	my ($now, $months, %options) = @_;
 
 	# months are 0..11
-	my ($j, $j, $j, $d, $m11, $y) = &righttime($now, %options);
+	my ($d, $m11, $y) = (&righttime($now, %options)) [ 3,4,5 ] ;
 
 	$y += 1900;
 
@@ -801,7 +908,7 @@ sub righttime
 
 sub parse_year_only
 {
-	my ($tr, $yr) = @_;
+	my ($tr, $yr, $now, %options) = @_;
 
 	$$tr =~ s#^\s+##;
 
@@ -810,10 +917,7 @@ sub parse_year_only
 		printf "matched at %d.\n", __LINE__ if $debug;
 		return 1;
 	} elsif ($$tr =~ s#\'(\d\d)(?:\s+|$ )##) {
-		$$yr = $1;
-		# This will stop working in year 2100.
-		$$yr += 100  if $yr < 70;
-		$$yr += 1900;
+		$$yr = expand_two_digit_year($1, $now, %options);
 		printf "matched at %d.\n", __LINE__ if $debug;
 		return 1;
 	}
@@ -969,6 +1073,9 @@ Date parsing can also use options.  The options are as follows:
 	TIMEFIRST -> try parsing time before date [not default]
 	PREFER_PAST -> when year or day of week is ambigueous, assume past
 	PREFER_FUTURE -> when year or day of week is ambigueous, assume future
+	SUBSECOND -> parse fraction seconds
+	VALIDATE -> only accept normal values for HHMMSS, YYMMDD.  Otherwise
+		days like -1 might give the last day of the previous month.
 
 =head1 DATE FORMATS RECOGNIZED
 
@@ -1015,7 +1122,7 @@ Date parsing can also use options.  The options are as follows:
 
 =head2 Absolute time formats:
 
-	hh:mm:ss 
+	hh:mm:ss[.ddd] 
 	hh:mm 
 	hh:mm[AP]M
 	hh[AP]M
@@ -1052,21 +1159,30 @@ date and a time are specified.  There are numerous options for
 controlling what is recognized and what is not.
 
 The return code is always the time in seconds since January 1st, 1970
-or zero if it was unable to parse the time.
+or undef if it was unable to parse the time.
 
 If a timezone is specified it must be after the time.  Year specifications
 can be tacked onto the end of absolute times.
 
+If C<parsedate()> is called from array contect, then it will return two
+elements.  On sucessful parses, it will return the seconds and what 
+remains of its input string.  On unsucessful parses, it will return
+C<undef> and an error string.
+
 =head1 EXAMPLES
 
 	$seconds = parsedate("Mon Jan  2 04:24:27 1995");
 	$seconds = parsedate("Tue Apr 4 00:22:12 PDT 1995");
 	$seconds = parsedate("04.04.95 00:22", ZONE => PDT);
+	$seconds = parsedate("Jan 1 1999 11:23:34.578", SUBSECOND => 1);
 	$seconds = parsedate("122212 950404", ZONE => PDT, TIMEFIRST => 1);
 	$seconds = parsedate("+3 secs", NOW => 796978800);
 	$seconds = parsedate("2 months", NOW => 796720932);
 	$seconds = parsedate("last Tuesday");
 
+	($seconds, $remaining) = parsedate("today is the day");
+	($seconds, $error) = parsedate("today is", WHOLE=>1);
+
 =head1 AUTHOR
 
 David Muir Sharnoff <muir@idiom.com>
